"""
Enterprise Integrated Channel Cornerstone Django application initialization.
"""

from django.apps import AppConfig


class CornerstoneConfig(AppConfig):
    """
    Configuration for the Enterprise Integrated Channel Cornerstone Django application.
    """
    name = 'integrated_channels.cornerstone'
    verbose_name = "Enterprise Cornerstone Integration"



"""
Class for transmitting content metadata to SuccessFactors.
"""
from integrated_channels.integrated_channel.transmitters.content_metadata import ContentMetadataTransmitter
from integrated_channels.sap_success_factors.client import SAPSuccessFactorsAPIClient


class SapSuccessFactorsContentMetadataTransmitter(ContentMetadataTransmitter):
    """
    This transmitter transmits exported content metadata to SAPSF.
    """

    def __init__(self, enterprise_configuration, client=SAPSuccessFactorsAPIClient):
        """
        Use the ``SAPSuccessFactorsAPIClient`` for content metadata transmission to SAPSF.
        """
        super().__init__(
            enterprise_configuration=enterprise_configuration,
            client=client
        )

    def transmit(self, create_payload, update_payload, delete_payload):
        """
        Transmit method overriding base transmissions. Due to rate limiting on SAP
        that prevents us from calling the post endpoint multiple times in one run,
        we prioritize the transmissions in DELETE, UPDATE, CREATE order
        """

        delete_payload_results, create_payload_results, update_payload_results = {}, {}, {}

        if delete_payload:
            self._log_info_for_each_item_map(delete_payload, 'transmitting delete')
            delete_payload_results = self._transmit_delete(delete_payload)
        elif update_payload:
            self._log_info_for_each_item_map(update_payload, 'transmitting update')
            update_payload_results = self._transmit_update(update_payload)
        elif create_payload:
            self._log_info_for_each_item_map(create_payload, 'transmitting create')
            create_payload_results = self._transmit_create(create_payload)
        return create_payload_results, update_payload_results, delete_payload_results

    def _prepare_items_for_transmission(self, channel_metadata_items):
        return {
            'ocnCourses': channel_metadata_items
        }



"""
Views for the ``pending-enterprise-customer-user`` API endpoint.
"""

from django_filters.rest_framework import DjangoFilterBackend
from edx_rbac.decorators import permission_required
from rest_framework import filters, permissions, status
from rest_framework.decorators import action
from rest_framework.response import Response
from rest_framework.status import HTTP_400_BAD_REQUEST

from enterprise import models
from enterprise.api.v1 import serializers
from enterprise.api.v1.views.base_views import EnterpriseReadWriteModelViewSet
from enterprise.logging import getEnterpriseLogger

LOGGER = getEnterpriseLogger(__name__)


class PendingEnterpriseCustomerUserViewSet(EnterpriseReadWriteModelViewSet):
    """
    API views for the ``pending-enterprise-learner`` API endpoint.
    Requires staff permissions
    """
    queryset = models.PendingEnterpriseCustomerUser.objects.all()
    filter_backends = (filters.OrderingFilter, DjangoFilterBackend)
    serializer_class = serializers.PendingEnterpriseCustomerUserSerializer
    permission_classes = (permissions.IsAuthenticated, permissions.IsAdminUser)

    FIELDS = (
        'enterprise_customer', 'user_email',
    )
    filterset_fields = FIELDS
    ordering_fields = FIELDS

    UNIQUE = 'unique'
    USER_EXISTS_ERROR = 'EnterpriseCustomerUser record already exists'

    def _get_return_status(self, serializer, many):
        """
        Run serializer validation and get return status
        """
        return_status = None
        serializer.is_valid(raise_exception=True)
        if not many:
            _, created = serializer.save()
            return_status = status.HTTP_201_CREATED if created else status.HTTP_204_NO_CONTENT
            return return_status

        data_list = serializer.save()
        for _, created in data_list:
            if created:
                return status.HTTP_201_CREATED
        return status.HTTP_204_NO_CONTENT

    def create(self, request, *args, **kwargs):
        """
        Creates a PendingEnterpriseCustomerUser if no EnterpriseCustomerUser for the given (customer, email)
        combination(s) exists.
        Can accept one user or a list of users.

        Returns 201 if any users were created, 204 if no users were created.
        """
        serializer = self.get_serializer(data=request.data, many=isinstance(request.data, list))
        return_status = self._get_return_status(serializer, many=isinstance(request.data, list))
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=return_status, headers=headers)


class PendingEnterpriseCustomerUserEnterpriseAdminViewSet(PendingEnterpriseCustomerUserViewSet):
    """
    Viewset for allowing enterprise admins to create linked learners
    Endpoint url: link_pending_enterprise_users/(?P<enterprise_uuid>[A-Za-z0-9-]+)/?$
    Admin must be an administrator for the enterprise in question
    """
    permission_classes = (permissions.IsAuthenticated,)
    serializer_class = serializers.LinkLearnersSerializer

    @action(methods=['post'], detail=False)
    @permission_required('enterprise.can_access_admin_dashboard', fn=lambda request, enterprise_uuid: enterprise_uuid)
    def link_learners(self, request, enterprise_uuid):
        """
        Creates a PendingEnterpriseCustomerUser if no EnterpriseCustomerUser for the given (customer, email)
        combination(s) exists.
        Can accept one user or a list of users.

        Returns 201 if any users were created, 204 if no users were created.
        """
        if not request.data:
            LOGGER.error('Empty user email payload in link_learners for enterprise: %s', enterprise_uuid)
            return Response(
                'At least one user email is required.',
                status=HTTP_400_BAD_REQUEST,
            )
        context = {'enterprise_customer__uuid': enterprise_uuid}
        serializer = self.get_serializer(
            data=request.data,
            many=isinstance(request.data, list),
            context=context,
        )
        return_status = self._get_return_status(serializer, many=isinstance(request.data, list))
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=return_status, headers=headers)



"""
Pagination helpers for enterprise api.
"""

from collections import OrderedDict
from urllib.parse import urlparse

from edx_rest_framework_extensions.paginators import DefaultPagination
from rest_framework.response import Response

from enterprise.toggles import enterprise_features


class PaginationWithFeatureFlags(DefaultPagination):
    """
    Adds a ``features`` dictionary to the default paginated response
    provided by edx_rest_framework_extensions. The ``features`` dict
    represents a collection of Waffle-based feature flags/samples/switches
    that may be used to control whether certain aspects of the system are
    enabled or disabled (e.g., feature flag turned on for all staff users but
    not turned on for real customers/learners).
    """

    def get_paginated_response(self, data):
        """
        Modifies the default paginated response to include ``enterprise_features`` dict.

        Arguments:
            self: PaginationWithFeatureFlags instance.
            data (dict): Results for current page.

        Returns:
            (Response): DRF response object containing ``enterprise_features`` dict.
        """
        paginated_response = super().get_paginated_response(data)
        paginated_response.data.update({
            'enterprise_features': enterprise_features(),
        })
        return paginated_response


def get_paginated_response(data, request):
    """
    Update pagination links in course catalog data and return DRF Response.

    Arguments:
        data (dict): Dictionary containing catalog courses.
        request (HttpRequest): Current request object.

    Returns:
        (Response): DRF response object containing pagination links.
    """
    url = urlparse(request.build_absolute_uri())._replace(query=None).geturl()

    next_page = None
    previous_page = None

    if data['next']:
        next_page = "{base_url}?{query_parameters}".format(
            base_url=url,
            query_parameters=urlparse(data['next']).query,
        )
        next_page = next_page.rstrip('?')
    if data['previous']:
        previous_page = "{base_url}?{query_parameters}".format(
            base_url=url,
            query_parameters=urlparse(data['previous'] or "").query,
        )
        previous_page = previous_page.rstrip('?')

    return Response(OrderedDict([
        ('count', data['count']),
        ('next', next_page),
        ('previous', previous_page),
        ('results', data['results'])
    ]))



"""
URL definitions for enterprise_learner_portal API endpoint.
"""

from django.urls import path

from enterprise_learner_portal.api.v1.views import EnterpriseCourseEnrollmentView

urlpatterns = [
    path('enterprise_course_enrollments/', EnterpriseCourseEnrollmentView.as_view(),
         name="enterprise-learner-portal-course-enrollment-list"
         ),
]



# Generated by Django 2.2.19 on 2021-11-29 21:14

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0147_auto_20211129_1949'),
    ]

    operations = [
        migrations.AlterField(
            model_name='enterprisecustomer',
            name='enable_universal_link',
            field=models.BooleanField(default=False, help_text='Specifies whether universal link generation is enabled for customer. Managed via admin portal settings UI.'),
        ),
        migrations.AlterField(
            model_name='historicalenterprisecustomer',
            name='enable_universal_link',
            field=models.BooleanField(default=False, help_text='Specifies whether universal link generation is enabled for customer. Managed via admin portal settings UI.'),
        ),
    ]



"""
Views for the Admin Notification API.
"""


from edx_rbac.decorators import permission_required
from edx_rest_framework_extensions.auth.jwt.authentication import JwtAuthentication
from rest_framework import permissions
from rest_framework.authentication import SessionAuthentication
from rest_framework.response import Response
from rest_framework.status import HTTP_200_OK, HTTP_400_BAD_REQUEST, HTTP_500_INTERNAL_SERVER_ERROR
from rest_framework.views import APIView

from enterprise import models
from enterprise.api.throttles import ServiceUserThrottle
from enterprise.errors import AdminNotificationAPIRequestError
from enterprise.logging import getEnterpriseLogger
from enterprise.utils import get_request_value

LOGGER = getEnterpriseLogger(__name__)


class NotificationReadView(APIView):
    """
    API to mark notifications as read.
    """
    permission_classes = (permissions.IsAuthenticated,)
    authentication_classes = (JwtAuthentication, SessionAuthentication,)
    throttle_classes = (ServiceUserThrottle,)

    REQUIRED_PARAM_NOTIFICATION_ID = 'notification_id'
    REQUIRED_PARAM_ENTERPRISE_SLUG = 'enterprise_slug'

    MISSING_REQUIRED_PARAMS_MSG = 'Some required parameter(s) missing: {}'

    def get_required_query_params(self, request):
        """
        Gets ``notification_id`` and ``enterprise_slug``.
        which are the relevant parameters for this API endpoint.

        :param request: The request to this endpoint.
        :return: The ``notification_id`` and ``enterprise_slug`` from the request.
        """
        enterprise_slug = get_request_value(request, self.REQUIRED_PARAM_ENTERPRISE_SLUG, '')
        notification_id = get_request_value(request, self.REQUIRED_PARAM_NOTIFICATION_ID, '')
        if not (notification_id and enterprise_slug):
            raise AdminNotificationAPIRequestError(
                self.get_missing_params_message([
                    (self.REQUIRED_PARAM_NOTIFICATION_ID, bool(notification_id)),
                    (self.REQUIRED_PARAM_ENTERPRISE_SLUG, bool(enterprise_slug)),
                ])
            )
        return notification_id, enterprise_slug

    def get_missing_params_message(self, parameter_state):
        """
        Get a user-friendly message indicating a missing parameter for the API endpoint.
        """
        params = ', '.join(name for name, present in parameter_state if not present)
        return self.MISSING_REQUIRED_PARAMS_MSG.format(params)

    @permission_required('enterprise.can_access_admin_dashboard')
    def post(self, request):
        """
        POST /enterprise/api/v1/read_notification

        Requires a JSON object of the following format::

            {
                'notification_id': 1,
                'enterprise_slug': 'enterprise_slug',
            }

        Keys:
            notification_id: Notification ID which is read by Current User.
            enterprise_slug: The slug of the enterprise.
        """
        try:
            notification_id, enterprise_slug = self.get_required_query_params(request)
        except AdminNotificationAPIRequestError as invalid_request:
            return Response({'error': str(invalid_request)}, status=HTTP_400_BAD_REQUEST)

        try:
            data = {
                self.REQUIRED_PARAM_NOTIFICATION_ID: notification_id,
                self.REQUIRED_PARAM_ENTERPRISE_SLUG: enterprise_slug,
            }
            enterprise_customer_user = models.EnterpriseCustomerUser.objects.get(
                enterprise_customer__slug=enterprise_slug, user_id=request.user.id
            )
            notification_read, _ = models.AdminNotificationRead.objects.get_or_create(
                enterprise_customer_user=enterprise_customer_user,
                admin_notification_id=notification_id,
                is_read=True
            )
            LOGGER.info(
                '[Admin Notification API] Notification read request successful. AdminNotificationRead ID'
                ' {}.'.format(notification_read.id)
            )
            return Response(data, status=HTTP_200_OK)
        except Exception as exc:  # pylint: disable=broad-except
            LOGGER.error(
                '[Admin Notification API] Notification read request failed, AdminNotification ID:{},Enterprise Slug:{}'
                ' User ID:{}, Exception:{}.'.format(notification_id, enterprise_slug, request.user.id, exc)
            )
            return Response(
                {'error': 'Notification read request failed'},
                status=HTTP_500_INTERNAL_SERVER_ERROR
            )



# Generated by Django 3.2.17 on 2023-02-21 18:35

from django.db import migrations, models
import uuid


def create_license_enrollment_uuid(apps, schema_editor):
    licensed_enrollments = apps.get_model('enterprise', 'LicensedEnterpriseCourseEnrollment')
    for licensed_enrollment in licensed_enrollments.objects.all():
        licensed_enrollment.uuid = uuid.uuid4()
        licensed_enrollment.save()


def create_learner_credit_enrollment_uuid(apps, schema_editor):
    learner_credit_enrollments = apps.get_model('enterprise', 'learnercreditenterprisecourseenrollment')
    for learner_credit_enrollment in learner_credit_enrollments.objects.all():
        learner_credit_enrollment.uuid = uuid.uuid4()
        learner_credit_enrollment.save()


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', '0167_auto_20230209_2108'),
    ]

    operations = [
        migrations.RunPython(create_license_enrollment_uuid, reverse_code=migrations.RunPython.noop),
        migrations.RunPython(create_learner_credit_enrollment_uuid, reverse_code=migrations.RunPython.noop),
    ]



"""
Write views for the ``enterprise-customer-catalog`` API endpoint.
"""
from urllib.parse import quote_plus, unquote

from edx_rbac.decorators import permission_required
from rest_framework import permissions, status
from rest_framework.decorators import action
from rest_framework.renderers import JSONRenderer
from rest_framework.response import Response
from rest_framework_xml.renderers import XMLRenderer

from django.conf import settings
from django.http import Http404
from django.utils.decorators import method_decorator
from django.utils.translation import gettext as _

from enterprise import models
from enterprise.api.utils import get_enterprise_customer_from_catalog_id
from enterprise.api.v1 import serializers
from enterprise.api.v1.decorators import require_at_least_one_query_parameter
from enterprise.api.v1.views.base_views import EnterpriseReadOnlyModelViewSet, EnterpriseWriteOnlyModelViewSet
from enterprise.constants import COURSE_KEY_URL_PATTERN
from enterprise.logging import getEnterpriseLogger

LOGGER = getEnterpriseLogger(__name__)


class EnterpriseCustomerCatalogWriteViewSet(EnterpriseWriteOnlyModelViewSet):
    """
    API write only views for the ``enterprise-customer-catalog`` API endpoint.
    """
    queryset = models.EnterpriseCustomerCatalog.objects.all()
    permission_classes = (permissions.IsAdminUser,)
    serializer_class = serializers.EnterpriseCustomerCatalogWriteOnlySerializer

    def has_enterprise_customer_catalog(self, uuid):
        try:
            return models.EnterpriseCustomerCatalog.objects.get(pk=uuid)
        except models.EnterpriseCustomerCatalog.DoesNotExist:
            return False

    def create(self, request, *args, **kwargs):
        """
        Creates a new EnterpriseCustomerCatalog and returns the created object.

        If an EnterpriseCustomerCatalog already exists for the given enterprise_customer and enterprise_catalog_query,
        returns the existing object.

        URL: /enterprise/api/v1/enterprise-customer-catalog/

        Method: POST

        Payload::

          {
            "title":  string - Title of the catalog,
            "enterprise_customer": string - UUID of an existing enterprise customer,
            "enterprise_catalog_query": string - id of an existing enterprise catalog query,
          }

        Returns 201 if a new EnterpriseCustomerCatalog was created, 200 if an existing EnterpriseCustomerCatalog was
        """

        enterprise_customer_uuid = request.data.get('enterprise_customer')
        enterprise_catalog_query_id = request.data.get('enterprise_catalog_query')
        enterprise_customer_catalog_list = models.EnterpriseCustomerCatalog.objects.filter(
            enterprise_customer=enterprise_customer_uuid)
        for catalog in enterprise_customer_catalog_list:
            catalog_query = catalog.enterprise_catalog_query
            if catalog_query is not None and catalog_query.id == int(enterprise_catalog_query_id):
                serialized_customer_catalog = serializers.EnterpriseCustomerCatalogWriteOnlySerializer(
                    catalog)
                LOGGER.info(
                    'EnterpriseCustomerCatalog already exists for enterprise_customer_uuid: %s '
                    'and enterprise_catalog_query_id: %s, using existing catalog: %s',
                    enterprise_customer_uuid, enterprise_catalog_query_id, catalog.uuid)
                return Response(serialized_customer_catalog.data, status=status.HTTP_200_OK)
        LOGGER.info(
            'Creating new EnterpriseCustomerCatalog for enterprise_customer_uuid: %s '
            'and enterprise_catalog_query_id: %s',
            enterprise_customer_uuid, enterprise_catalog_query_id)
        return super().create(request, *args, **kwargs)

    def partial_update(self, request, *args, **kwargs):
        """
        Partially updates only the title in EnterpriseCustomerCatalog and returns the updated object.

        URL: /enterprise/api/v1/enterprise-customer-catalog/

        Method: PATCH

        Payload::

          {
            "uuid": string - UUID of an existing enterprise customer catalog,
            "title":  string - Title of the catalog,
          }

        Returns 200 along with the updated object.
        """

        enterprise_customer_catalog_uuid = request.data.get('uuid')
        found_catalog = self.has_enterprise_customer_catalog(enterprise_customer_catalog_uuid)
        if not found_catalog:
            return Response(
                {'detail': f'Could not find catalog uuid {enterprise_customer_catalog_uuid}'},
                status=status.HTTP_404_NOT_FOUND
            )

        serializer = self.get_serializer(found_catalog, data=request.data, partial=True)
        if serializer.is_valid(raise_exception=True):
            serializer.save()
        return Response(serializer.data, status=status.HTTP_200_OK)


class EnterpriseCustomerCatalogViewSet(EnterpriseReadOnlyModelViewSet):
    """
    API Views for performing search through course discovery at the ``enterprise_catalogs`` API endpoint.
    """
    queryset = models.EnterpriseCustomerCatalog.objects.all()

    USER_ID_FILTER = 'enterprise_customer__enterprise_customer_users__user_id'
    FIELDS = (
        'uuid', 'enterprise_customer',
    )
    filterset_fields = FIELDS
    ordering_fields = FIELDS
    renderer_classes = (JSONRenderer, XMLRenderer,)

    @permission_required('enterprise.can_view_catalog', fn=lambda request, *args, **kwargs: None)
    def list(self, request, *args, **kwargs):
        return super().list(request, *args, **kwargs)

    @permission_required(
        'enterprise.can_view_catalog',
        fn=lambda request, *args, **kwargs: get_enterprise_customer_from_catalog_id(kwargs['pk']))
    def retrieve(self, request, *args, **kwargs):
        return super().retrieve(request, *args, **kwargs)

    def get_serializer_class(self):
        view_action = getattr(self, 'action', None)
        if view_action == 'retrieve':
            return serializers.EnterpriseCustomerCatalogDetailSerializer
        return serializers.EnterpriseCustomerCatalogSerializer

    @method_decorator(require_at_least_one_query_parameter('course_run_ids', 'program_uuids'))
    @action(detail=True)
    # pylint: disable=unused-argument
    def contains_content_items(self, request, pk, course_run_ids, program_uuids):
        """
        Return whether or not the EnterpriseCustomerCatalog contains the specified content.

        Multiple course_run_ids and/or program_uuids query parameters can be sent to this view to check
        for their existence in the EnterpriseCustomerCatalog. At least one course run key
        or program UUID value must be included in the request.
        """
        enterprise_customer_catalog = self.get_object()

        # Maintain plus characters in course key.
        course_run_ids = [unquote(quote_plus(course_run_id)) for course_run_id in course_run_ids]

        contains_content_items = True
        if course_run_ids:
            contains_content_items = enterprise_customer_catalog.contains_courses(course_run_ids)
        if program_uuids:
            contains_content_items = (
                contains_content_items and
                enterprise_customer_catalog.contains_programs(program_uuids)
            )

        return Response({'contains_content_items': contains_content_items})

    @action(detail=True, url_path='courses/{}'.format(COURSE_KEY_URL_PATTERN))
    @permission_required(
        'enterprise.can_view_catalog',
        fn=lambda request, pk, course_key: get_enterprise_customer_from_catalog_id(pk))
    def course_detail(self, request, pk, course_key):  # pylint: disable=unused-argument
        """
        Return the metadata for the specified course.

        The course needs to be included in the specified EnterpriseCustomerCatalog
        in order for metadata to be returned from this endpoint.
        """
        enterprise_customer_catalog = self.get_object()
        course = enterprise_customer_catalog.get_course(course_key)
        if not course:
            error_message = _(
                '[Enterprise API] CourseKey not found in the Catalog. Course: {course_key}, Catalog: {catalog_id}'
            ).format(
                course_key=course_key,
                catalog_id=enterprise_customer_catalog.uuid,
            )
            LOGGER.warning(error_message)
            raise Http404

        context = self.get_serializer_context()
        context['enterprise_customer_catalog'] = enterprise_customer_catalog
        serializer = serializers.CourseDetailSerializer(course, context=context)
        return Response(serializer.data)

    @action(detail=True, url_path='course_runs/{}'.format(settings.COURSE_ID_PATTERN))
    @permission_required(
        'enterprise.can_view_catalog',
        fn=lambda request, pk, course_id: get_enterprise_customer_from_catalog_id(pk))
    def course_run_detail(self, request, pk, course_id):  # pylint: disable=unused-argument
        """
        Return the metadata for the specified course run.

        The course run needs to be included in the specified EnterpriseCustomerCatalog
        in order for metadata to be returned from this endpoint.
        """
        enterprise_customer_catalog = self.get_object()
        course_run = enterprise_customer_catalog.get_course_run(course_id)
        if not course_run:
            error_message = _(
                '[Enterprise API] CourseRun not found in the Catalog. CourseRun: {course_id}, Catalog: {catalog_id}'
            ).format(
                course_id=course_id,
                catalog_id=enterprise_customer_catalog.uuid,
            )
            LOGGER.warning(error_message)
            raise Http404

        context = self.get_serializer_context()
        context['enterprise_customer_catalog'] = enterprise_customer_catalog
        serializer = serializers.CourseRunDetailSerializer(course_run, context=context)
        return Response(serializer.data)

    @action(detail=True, url_path='programs/(?P<program_uuid>[^/]+)')
    @permission_required(
        'enterprise.can_view_catalog',
        fn=lambda request, pk, program_uuid: get_enterprise_customer_from_catalog_id(pk))
    def program_detail(self, request, pk, program_uuid):  # pylint: disable=unused-argument
        """
        Return the metadata for the specified program.

        The program needs to be included in the specified EnterpriseCustomerCatalog
        in order for metadata to be returned from this endpoint.
        """
        enterprise_customer_catalog = self.get_object()
        program = enterprise_customer_catalog.get_program(program_uuid)
        if not program:
            error_message = _(
                '[Enterprise API] Program not found in the Catalog. Program: {program_uuid}, Catalog: {catalog_id}'
            ).format(
                program_uuid=program_uuid,
                catalog_id=enterprise_customer_catalog.uuid,
            )
            LOGGER.warning(error_message)
            raise Http404

        context = self.get_serializer_context()
        context['enterprise_customer_catalog'] = enterprise_customer_catalog
        serializer = serializers.ProgramDetailSerializer(program, context=context)
        return Response(serializer.data)



"""
Api client for the SSO Orchestrator API.
"""
from urllib.parse import urljoin

import requests
from edx_rest_api_client.client import get_request_id, user_agent
from requests.auth import HTTPBasicAuth
from requests.exceptions import RequestException
from rest_framework.reverse import reverse

from django.conf import settings

from enterprise.utils import (
    get_configuration_value,
    get_sso_orchestrator_api_base_url,
    get_sso_orchestrator_basic_auth_password,
    get_sso_orchestrator_basic_auth_username,
    get_sso_orchestrator_configure_path,
)

USER_AGENT = user_agent()


class SsoOrchestratorClientError(RequestException):
    """
    Indicate a problem when interacting with an enterprise api client.
    """


class EnterpriseSSOOrchestratorApiClient:
    """
    The enterprise API client to communicate with the SSO Orchestrator API. Reads conf settings values to determine
    orchestration paths and credentials.

    Required settings values:
    - LMS_ROOT_URL
    - SSO_ORCHESTRATOR_API_BASE_URL
    - SSO_ORCHESTRATOR_CONFIGURE_PATH
    - SSO_ORCHESTRATOR_BASIC_AUTH_USERNAME
    - SSO_ORCHESTRATOR_BASIC_AUTH_PASSWORD
    """

    def __init__(self):
        self.base_url = get_sso_orchestrator_api_base_url()
        self.session = None

    def _get_orchestrator_callback_url(self, config_pk):
        """
        get the callback url for the SSO Orchestrator API
        """
        lms_base_url = get_configuration_value('LMS_ROOT_URL', settings.LMS_ROOT_URL)
        if not lms_base_url:
            raise SsoOrchestratorClientError(
                "Failed to create SSO Orchestrator callback url: LMS_ROOT_URL required.",
            )
        path = reverse(
            'enterprise-customer-sso-configuration-orchestration-complete',
            kwargs={'configuration_uuid': config_pk},
        )
        return urljoin(lms_base_url, path)

    def _get_orchestrator_configure_url(self):
        """
        get the configure url for the SSO Orchestrator API
        """
        # probably want config value validated for this
        return urljoin(self.base_url, get_sso_orchestrator_configure_path())

    def _create_auth_header(self):
        """
        create the basic auth header for requests to the SSO Orchestrator API
        """
        if orchestrator_username := get_sso_orchestrator_basic_auth_username():
            if orchestrator_password := get_sso_orchestrator_basic_auth_password():
                return HTTPBasicAuth(orchestrator_username, orchestrator_password)
            else:
                raise SsoOrchestratorClientError(
                    "Failed to create SSO Orchestrator auth headers: password required.",
                )
        raise SsoOrchestratorClientError(
            "Failed to create SSO Orchestrator auth headers: username required.",
        )

    def _create_session(self):
        """
        create a requests session object
        """
        if not self.session:
            self.session = requests.Session()
            self.session.headers['User-Agent'] = USER_AGENT
            self.session.headers['X-Request-ID'] = get_request_id()

    def _post(self, url, data=None):
        """
        make a GET request to the SSO Orchestrator API
        """
        self._create_session()
        response = self.session.post(url, json=data, auth=self._create_auth_header())
        if response.status_code >= 300:
            raise SsoOrchestratorClientError(
                f"Failed to make SSO Orchestrator API request: {response.status_code}",
                response=response,
            )
        return response.status_code

    def configure_sso_orchestration_record(
        self,
        config_data,
        config_pk,
        enterprise_data,
        is_sap=False,
        updating_existing_record=False,
        sap_config_data=None,
    ):
        """
        configure an SSO orchestration record
        """
        config_data['uuid'] = str(config_data['uuid'])
        request_data = {
            'samlConfiguration': config_data,
            'requestIdentifier': str(config_pk),
            'enterprise': enterprise_data,
        }

        callback_url = self._get_orchestrator_callback_url(str(config_pk))
        if updating_existing_record:
            callback_url = f"{callback_url}?updating_existing_record=true"
        request_data['callbackUrl'] = callback_url

        if is_sap or sap_config_data:
            request_data['sapsfConfiguration'] = sap_config_data

        return self._post(self._get_orchestrator_configure_url(), data=request_data)



# Generated by Django 3.2.20 on 2023-08-24 17:49

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import model_utils.fields
import simple_history.models
import uuid


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('enterprise', '0180_chatgptresponse'),
    ]

    operations = [
        migrations.CreateModel(
            name='HistoricalEnterpriseCustomerSsoConfiguration',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('is_removed', models.BooleanField(default=False)),
                ('uuid', models.UUIDField(db_index=True, default=uuid.uuid4, editable=False)),
                ('active', models.BooleanField(default=False, help_text='Specifies whether the integration with the SSO orchestration is active.')),
                ('identity_provider', models.CharField(help_text='The identity provider integrated with by the SSO orchestrator .', max_length=255)),
                ('metadata_url', models.CharField(help_text='The metadata url of the identity provider.', max_length=255)),
                ('metadata_xml', models.TextField(blank=True, help_text='The metadata xml of the identity provider.', null=True)),
                ('entity_id', models.CharField(help_text='The entity id of the identity provider.', max_length=255)),
                ('update_from_metadata', models.BooleanField(default=True, help_text="Specifies whether the integration with the customer's identity provider should auto update metadata.")),
                ('user_id_attribute', models.CharField(blank=True, max_length=128, null=True)),
                ('full_name_attribute', models.CharField(blank=True, max_length=128, null=True)),
                ('last_name_attribute', models.CharField(blank=True, max_length=128, null=True)),
                ('email_attribute', models.CharField(blank=True, max_length=128, null=True)),
                ('username_attribute', models.CharField(blank=True, max_length=128, null=True)),
                ('country_attribute', models.CharField(blank=True, max_length=128, null=True)),
                ('submitted_at', models.DateTimeField(blank=True, help_text='The date and time when the configuration was submitted to the SSO orchestration.', null=True)),
                ('configured_at', models.DateTimeField(blank=True, help_text='The date and time when the configuration was completed by the SSO orchestration.', null=True)),
                ('validated_at', models.DateTimeField(blank=True, help_text='The date and time when the configuration was validated and used for the first time.', null=True)),
                ('odata_api_timeout_interval', models.PositiveIntegerField(blank=True, default=29, help_text='SAP specific configuration: the timeout interval for the OData API in seconds.', null=True)),
                ('odata_api_root_url', models.CharField(blank=True, help_text='SAP specific configuration: the root url of the OData API.', max_length=255, null=True)),
                ('odata_company_id', models.CharField(blank=True, help_text='SAP specific configuration: the company id of the OData API.', max_length=255, null=True)),
                ('sapsf_oauth_root_url', models.CharField(blank=True, help_text='SAP specific configuration: the root url of the SAP SuccessFactors OAuth API.', max_length=255, null=True)),
                ('odata_api_request_timeout', models.PositiveIntegerField(blank=True, default=29, help_text='SAP specific configuration: the timeout interval for the OData API in seconds.', null=True)),
                ('sapsf_private_key', models.TextField(blank=True, help_text='SAP specific configuration: the private key used to sign the SAML assertion.', null=True)),
                ('odata_client_id', models.CharField(blank=True, help_text='SAP specific configuration: the client id of the OData API.', max_length=255, null=True)),
                ('oauth_user_id', models.CharField(blank=True, help_text='SAP specific configuration: the user id of the OAuth API.', max_length=255, null=True)),
                ('history_id', models.AutoField(primary_key=True, serialize=False)),
                ('history_date', models.DateTimeField()),
                ('history_change_reason', models.CharField(max_length=100, null=True)),
                ('history_type', models.CharField(choices=[('+', 'Created'), ('~', 'Changed'), ('-', 'Deleted')], max_length=1)),
                ('enterprise_customer', models.ForeignKey(blank=True, db_constraint=False, help_text='The enterprise that can be linked using this key.', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='+', to='enterprise.enterprisecustomer')),
                ('history_user', models.ForeignKey(null=True, on_delete=django.db.models.deletion.SET_NULL, related_name='+', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name': 'historical Enterprise Customer SSO Configuration',
                'verbose_name_plural': 'historical Enterprise Customer SSO Configurations',
                'ordering': ('-history_date', '-history_id'),
                'get_latest_by': ('history_date', 'history_id'),
            },
            bases=(simple_history.models.HistoricalChanges, models.Model),
        ),
        migrations.CreateModel(
            name='EnterpriseCustomerSsoConfiguration',
            fields=[
                ('created', model_utils.fields.AutoCreatedField(default=django.utils.timezone.now, editable=False, verbose_name='created')),
                ('modified', model_utils.fields.AutoLastModifiedField(default=django.utils.timezone.now, editable=False, verbose_name='modified')),
                ('is_removed', models.BooleanField(default=False)),
                ('uuid', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('active', models.BooleanField(default=False, help_text='Specifies whether the integration with the SSO orchestration is active.')),
                ('identity_provider', models.CharField(help_text='The identity provider integrated with by the SSO orchestrator .', max_length=255)),
                ('metadata_url', models.CharField(help_text='The metadata url of the identity provider.', max_length=255)),
                ('metadata_xml', models.TextField(blank=True, help_text='The metadata xml of the identity provider.', null=True)),
                ('entity_id', models.CharField(help_text='The entity id of the identity provider.', max_length=255)),
                ('update_from_metadata', models.BooleanField(default=True, help_text="Specifies whether the integration with the customer's identity provider should auto update metadata.")),
                ('user_id_attribute', models.CharField(blank=True, max_length=128, null=True)),
                ('full_name_attribute', models.CharField(blank=True, max_length=128, null=True)),
                ('last_name_attribute', models.CharField(blank=True, max_length=128, null=True)),
                ('email_attribute', models.CharField(blank=True, max_length=128, null=True)),
                ('username_attribute', models.CharField(blank=True, max_length=128, null=True)),
                ('country_attribute', models.CharField(blank=True, max_length=128, null=True)),
                ('submitted_at', models.DateTimeField(blank=True, help_text='The date and time when the configuration was submitted to the SSO orchestration.', null=True)),
                ('configured_at', models.DateTimeField(blank=True, help_text='The date and time when the configuration was completed by the SSO orchestration.', null=True)),
                ('validated_at', models.DateTimeField(blank=True, help_text='The date and time when the configuration was validated and used for the first time.', null=True)),
                ('odata_api_timeout_interval', models.PositiveIntegerField(blank=True, default=29, help_text='SAP specific configuration: the timeout interval for the OData API in seconds.', null=True)),
                ('odata_api_root_url', models.CharField(blank=True, help_text='SAP specific configuration: the root url of the OData API.', max_length=255, null=True)),
                ('odata_company_id', models.CharField(blank=True, help_text='SAP specific configuration: the company id of the OData API.', max_length=255, null=True)),
                ('sapsf_oauth_root_url', models.CharField(blank=True, help_text='SAP specific configuration: the root url of the SAP SuccessFactors OAuth API.', max_length=255, null=True)),
                ('odata_api_request_timeout', models.PositiveIntegerField(blank=True, default=29, help_text='SAP specific configuration: the timeout interval for the OData API in seconds.', null=True)),
                ('sapsf_private_key', models.TextField(blank=True, help_text='SAP specific configuration: the private key used to sign the SAML assertion.', null=True)),
                ('odata_client_id', models.CharField(blank=True, help_text='SAP specific configuration: the client id of the OData API.', max_length=255, null=True)),
                ('oauth_user_id', models.CharField(blank=True, help_text='SAP specific configuration: the user id of the OAuth API.', max_length=255, null=True)),
                ('enterprise_customer', models.ForeignKey(help_text='The enterprise that can be linked using this key.', on_delete=django.db.models.deletion.CASCADE, related_name='sso_orchestration_records', to='enterprise.enterprisecustomer')),
            ],
            options={
                'verbose_name': 'Enterprise Customer SSO Configuration',
                'verbose_name_plural': 'Enterprise Customer SSO Configurations',
            },
        ),
        migrations.AlterModelManagers(
            name='enterprisecustomerssoconfiguration',
            managers=[
                ('all_objects', django.db.models.manager.Manager()),
            ],
        ),
    ]



# Generated by Django 2.2.24 on 2021-10-14 14:29

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('integrated_channel', '0005_auto_20211005_1052'),
    ]

    operations = [
        migrations.AddField(
            model_name='contentmetadataitemtransmission',
            name='deleted_at',
            field=models.DateTimeField(blank=True, help_text='Date when the content transmission was deleted', null=True),
        ),
    ]



# -*- coding: utf-8 -*-
"""
Enterprise Integrated Channel Degreed2 Django application initialization.
"""

from django.apps import AppConfig


class Degreed2Config(AppConfig):
    """
    Configuration for the Enterprise Integrated Channel Degreed2 Django application.
    """
    name = 'integrated_channels.degreed2'
    verbose_name = "Enterprise Degreed2 Integration"
    oauth_api_path = "/oauth/token"
    courses_api_path = "/api/v2/content/courses"
    completions_api_path = "/api/v2/completions"



# Generated by Django 3.2.12 on 2022-03-24 15:50

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('cornerstone', '0015_auto_20220302_2231'),
    ]

    operations = [
        migrations.AddField(
            model_name='cornerstonelearnerdatatransmissionaudit',
            name='enterprise_customer_uuid',
            field=models.UUIDField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='cornerstonelearnerdatatransmissionaudit',
            name='instructor_name',
            field=models.CharField(blank=True, max_length=255),
        ),
        migrations.AddField(
            model_name='cornerstonelearnerdatatransmissionaudit',
            name='plugin_configuration_id',
            field=models.PositiveIntegerField(blank=True, null=True),
        ),
        migrations.AddField(
            model_name='cornerstonelearnerdatatransmissionaudit',
            name='total_hours',
            field=models.FloatField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='cornerstonelearnerdatatransmissionaudit',
            name='completed_timestamp',
            field=models.DateTimeField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='cornerstonelearnerdatatransmissionaudit',
            name='course_completed',
            field=models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name='cornerstonelearnerdatatransmissionaudit',
            name='course_id',
            field=models.CharField(max_length=255),
        ),
    ]



"""
Content metadata exporter for Canvas
"""

from logging import getLogger

from integrated_channels.integrated_channel.exporters.content_metadata import ContentMetadataExporter

LOGGER = getLogger(__name__)
BLACKBOARD_COURSE_CONTENT_NAME = 'edX Course Details'


class BlackboardContentMetadataExporter(ContentMetadataExporter):
    """
        Blackboard implementation of ContentMetadataExporter.
        Note: courseId is not being exported here (instead done in client during content send)
    """
    DATA_TRANSFORM_MAPPING = {
        'externalId': 'key',
        'course_metadata': 'course_metadata',
        'course_content_metadata': 'course_content_metadata',
        'course_child_content_metadata': 'course_child_content_metadata',
    }

    DESCRIPTION_TEXT_TEMPLATE = "<a href={enrollment_url} target=_blank>Go to edX course page</a><br/>"
    LARGE_DESCRIPTION_TEXT_TEMPLATE = "<a href={enrollment_url} style='font-size:150%' target=_blank>" \
                                      "Go to edX course page</a><br/>"

    COURSE_TITLE_TEMPLATE = '<h1 style="font-size:xxx-large; margin-bottom:0; margin-top:0">{title}</h1>'

    COURSE_DESCRIPTION_TEMPLATE = '<p style="width:60%;">{description}</p>'

    COURSE_CONTENT_IMAGE_TEMPLATE = '<img src={image_url} width="30%" height="25%" border="40px"/>'

    COURSE_CONTENT_BODY_TEMPLATE = '<div><div style="display: inline-block">' \
                                   '{course_title}{large_description_text}<hr/></div>' \
                                   '<br/>{course_content_image}' \
                                   '<br/><br/><br/>{course_description}' \
                                   '<br/>{description_text}</div>'.format(
                                       course_title=COURSE_TITLE_TEMPLATE,
                                       large_description_text=LARGE_DESCRIPTION_TEXT_TEMPLATE,
                                       course_content_image=COURSE_CONTENT_IMAGE_TEMPLATE,
                                       course_description=COURSE_DESCRIPTION_TEMPLATE,
                                       description_text=DESCRIPTION_TEXT_TEMPLATE,
                                   )

    def transform_course_metadata(self, content_metadata_item):
        """
        Formats the metadata necessary to create a base course object in Blackboard
        """
        return {
            'name': content_metadata_item.get('title', None),
            'externalId': content_metadata_item.get('key', None),
            'description': self.DESCRIPTION_TEXT_TEMPLATE.format(
                enrollment_url=content_metadata_item.get('enrollment_url', None)
            )
        }

    def transform_course_content_metadata(self, content_metadata_item):  # pylint: disable=unused-argument
        """
        Formats the metadata necessary to create a course content object in Blackboard
        """
        return {
            'title': BLACKBOARD_COURSE_CONTENT_NAME,
            'position': 0,
            "contentHandler": {"id": "resource/x-bb-folder"}
        }

    def transform_course_child_content_metadata(self, content_metadata_item):
        """
        Formats the metadata necessary to create a course content object in Blackboard
        """
        title = content_metadata_item.get('title', None)
        return {
            'title': BLACKBOARD_COURSE_CONTENT_NAME,
            'availability': 'Yes',
            'contentHandler': {
                'id': 'resource/x-bb-document',
            },
            'body': self.COURSE_CONTENT_BODY_TEMPLATE.format(
                title=title,
                description=content_metadata_item.get('full_description', None),
                image_url=content_metadata_item.get('image_url', None),
                enrollment_url=content_metadata_item.get('enrollment_url', None)
            )
        }



# Generated by Django 2.2.17 on 2021-01-11 12:53

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('enterprise', 'unique_constraints_pending_users'),
    ]

    operations = [
        migrations.AddField(
            model_name='enterprisecustomer',
            name='sender_alias',
            field=models.CharField(blank=True, help_text='Specifies the sender alias which will be used in enterprise customer Assign/Remind/Revoke emails.', max_length=255, null=True),
        ),
        migrations.AddField(
            model_name='historicalenterprisecustomer',
            name='sender_alias',
            field=models.CharField(blank=True, help_text='Specifies the sender alias which will be used in enterprise customer Assign/Remind/Revoke emails.', max_length=255, null=True),
        ),
    ]



"""
Assist integrated channels with retrieving learner completion data.

Module contains resources for integrated pipelines to retrieve all the
grade and completion data for enrollments belonging to a particular
enterprise customer.
"""

from datetime import datetime
from logging import getLogger

from opaque_keys import InvalidKeyError
from requests.exceptions import HTTPError

from django.apps import apps
from django.contrib import auth
from django.utils import timezone
from django.utils.dateparse import parse_datetime

from integrated_channels.integrated_channel.channel_settings import ChannelSettingsMixin

try:
    from openedx.core.djangoapps.content.course_overviews.models import CourseOverview
except ImportError:
    CourseOverview = None
from consent.models import DataSharingConsent
from enterprise.api_client.lms import GradesApiClient
from enterprise.models import EnterpriseCourseEnrollment
from integrated_channels.catalog_service_utils import get_course_id_for_enrollment
from integrated_channels.integrated_channel.exporters import Exporter
from integrated_channels.lms_utils import (
    get_completion_summary,
    get_course_certificate,
    get_course_details,
    get_persistent_grade,
    get_single_user_grade,
)
from integrated_channels.utils import (
    generate_formatted_log,
    is_already_transmitted,
    is_course_completed,
    parse_datetime_to_epoch_millis,
)

LOGGER = getLogger(__name__)
User = auth.get_user_model()


class LearnerExporter(ChannelSettingsMixin, Exporter):
    """
    Base class for exporting learner completion data to integrated channels.
    """

    GRADE_AUDIT = 'Audit'
    GRADE_PASSING = 'Pass'
    GRADE_FAILING = 'Fail'
    GRADE_INCOMPLETE = 'In Progress'

    def __init__(self, user, enterprise_configuration):
        """
        Store the data needed to export the learner data to the integrated channel.

        Arguments:

        * ``user``: User instance with access to the Grades API for the Enterprise Customer's courses.
        * ``enterprise_configuration``: EnterpriseCustomerPluginConfiguration instance for the current channel.
        """
        # The Grades API and Certificates API clients require an OAuth2 access token,
        # so cache the client to allow the token to be reused. Cache other clients for
        # general reuse.
        self.grades_api = None
        self.certificates_api = None
        self.course_api = None
        self.course_enrollment_api = None

        super().__init__(user, enterprise_configuration)

    @property
    def grade_passing(self):
        """
        Returns the string used for a passing grade.
        """
        return self.GRADE_PASSING

    @property
    def grade_failing(self):
        """
        Returns the string used for a failing grade.
        """
        return self.GRADE_FAILING

    @property
    def grade_incomplete(self):
        """
        Returns the string used for an incomplete course grade.
        """
        return self.GRADE_INCOMPLETE

    @property
    def grade_audit(self):
        """
        Returns the string used for an audit course grade.
        """
        return self.GRADE_AUDIT

    def bulk_assessment_level_export(self):
        """
        Collect assessment level learner data for the ``EnterpriseCustomer`` where data sharing consent is granted.

        Yields a ``LearnerDataTransmissionAudit`` for each subsection in a course under an enrollment, containing:

        * ``enterprise_course_enrollment_id``: The id reference to the ``EnterpriseCourseEnrollment`` object.
        * ``course_id``: The string ID of the course under the enterprise enrollment.
        * ``subsection_id``: The string ID of the subsection within the course.
        * ``grade``: string grade recorded for the learner in the course.
        """
        enrollment_queryset = EnterpriseCourseEnrollment.objects.select_related(
            'enterprise_customer_user'
        ).filter(
            enterprise_customer_user__enterprise_customer=self.enterprise_customer,
            enterprise_customer_user__active=True,
        ).order_by('course_id')

        # Create a record of each subsection from every enterprise enrollment
        for enterprise_enrollment in enrollment_queryset:
            if not LearnerExporter.has_data_sharing_consent(enterprise_enrollment):
                continue

            assessment_grade_data = self._collect_assessment_grades_data(enterprise_enrollment)

            records = self.get_learner_assessment_data_records(
                enterprise_enrollment=enterprise_enrollment,
                assessment_grade_data=assessment_grade_data,
            )
            if records:
                # There are some cases where we won't receive a record from the above
                # method; right now, that should only happen if we have an Enterprise-linked
                # user for the integrated channel, and transmission of that user's
                # data requires an upstream user identifier that we don't have (due to a
                # failure of SSO or similar). In such a case, `get_learner_data_record`
                # would return None, and we'd simply skip yielding it here.
                yield from records

    def single_assessment_level_export(self, **kwargs):
        """
        Collect an assessment level learner data for the ``EnterpriseCustomer`` where data sharing consent is
        granted.

        Yields a ``LearnerDataTransmissionAudit`` for each subsection of the course that the learner is enrolled in,
        containing:

        * ``enterprise_course_enrollment_id``: The id reference to the ``EnterpriseCourseEnrollment`` object.
        * ``course_id``: The string ID of the course under the enterprise enrollment.
        * ``subsection_id``: The string ID of the subsection within the course.
        * ``grade``: string grade recorded for the learner in the course.
        * ``learner_to_transmit``: REQUIRED User object, representing the learner whose data is being exported.

        """
        lms_user_for_filter = kwargs.get('learner_to_transmit')
        TransmissionAudit = kwargs.get('TransmissionAudit', None)
        course_run_id = kwargs.get('course_run_id', None)
        grade = kwargs.get('grade', None)
        subsection_id = kwargs.get('subsection_id')
        enrollment_queryset = EnterpriseCourseEnrollment.objects.select_related(
            'enterprise_customer_user'
        ).filter(
            enterprise_customer_user__active=True,
            enterprise_customer_user__user_id=lms_user_for_filter.id,
            course_id=course_run_id,
        ).order_by('course_id')

        # We are transmitting for an enrollment, so grab just the one.
        enterprise_enrollment = enrollment_queryset.first()

        if not enterprise_enrollment:
            LOGGER.info(generate_formatted_log(
                self.enterprise_configuration.channel_code(),
                self.enterprise_configuration.enterprise_customer.uuid,
                lms_user_for_filter,
                course_run_id,
                'Either qualifying enrollments not found for learner, or, '
                'enterprise_customer_user record is inactive. Skipping transmit assessment grades.'
            ))
            return

        already_transmitted = is_already_transmitted(
            TransmissionAudit,
            enterprise_enrollment.id,
            self.enterprise_configuration.id,
            grade,
            subsection_id,
            detect_grade_updated=self.INCLUDE_GRADE_FOR_COMPLETION_AUDIT_CHECK,
        )

        if not (TransmissionAudit and already_transmitted) and LearnerExporter.has_data_sharing_consent(
                enterprise_enrollment):

            assessment_grade_data = self._collect_assessment_grades_data(enterprise_enrollment)

            records = self.get_learner_assessment_data_records(
                enterprise_enrollment=enterprise_enrollment,
                assessment_grade_data=assessment_grade_data,
            )
            if records:
                # There are some cases where we won't receive a record from the above
                # method; right now, that should only happen if we have an Enterprise-linked
                # user for the integrated channel, and transmission of that user's
                # data requires an upstream user identifier that we don't have (due to a
                # failure of SSO or similar). In such a case, `get_learner_data_record`
                # would return None, and we'd simply skip yielding it here.
                yield from records

    @staticmethod
    def has_data_sharing_consent(enterprise_enrollment):
        """
        Helper method to determine whether an enrollment has data sharing consent or not.
        """
        consent = DataSharingConsent.objects.proxied_get(
            username=enterprise_enrollment.enterprise_customer_user.username,
            course_id=enterprise_enrollment.course_id,
            enterprise_customer=enterprise_enrollment.enterprise_customer_user.enterprise_customer
        )
        if consent.granted and not enterprise_enrollment.audit_reporting_disabled:
            return True

        return False

    def _determine_enrollments_permitted(
            self,
            lms_user_for_filter,
            course_run_id,
            channel_name,
            skip_transmitted,
            TransmissionAudit,
            grade,
    ):
        """
        Determines which enrollments can be safely transmitted after checking
        * enrollments that are already transmitted
        * enrollments that are permitted to be transmitted by selecting enrollments for which:
        *    - data sharing consent is granted
        *    - audit_reporting is enabled (via enterprise level switch)
        """
        enrollments_to_process = self.get_enrollments_to_process(
            lms_user_for_filter,
            course_run_id,
            channel_name,
        )

        if TransmissionAudit and skip_transmitted:
            untransmitted_enrollments = self._filter_out_pre_transmitted_enrollments(
                enrollments_to_process,
                channel_name,
                grade,
                TransmissionAudit
            )
        else:
            untransmitted_enrollments = enrollments_to_process

        # filter out enrollments which don't allow integrated_channels grade transmit
        enrollments_permitted = set()
        for enrollment in untransmitted_enrollments:
            if (not LearnerExporter.has_data_sharing_consent(enrollment) or
                    enrollment.audit_reporting_disabled):
                continue
            enrollments_permitted.add(enrollment)
        return enrollments_permitted

    def export_unique_courses(self):
        """
        Retrieve and export all unique course ID's from an enterprise customer's learner enrollments.
        """
        enrollment_queryset = EnterpriseCourseEnrollment.objects.select_related(
            'enterprise_customer_user'
        ).filter(
            enterprise_customer_user__enterprise_customer=self.enterprise_customer,
            enterprise_customer_user__active=True,
        ).order_by('course_id')
        return {get_course_id_for_enrollment(enrollment) for enrollment in enrollment_queryset}

    def get_grades_summary(
        self,
        course_details,
        enterprise_enrollment,
        channel_name,
        incomplete_count=None,
    ):
        '''
        Fetch grades info using either certificate api, or grades api.
        Note: This logic is going to be refactored, so that audit enrollments are treated separately
        - For audit enrollments, currently will fetch using grades api,
        - For non audit, it still calls grades api if pacing !=instructor otherwise calls certificate api
        This pacing logic needs cleanup for a more accurate piece of logic since pacing should not be relevant

        Returns: tuple with values:
            completed_date_from_api, grade_from_api, is_passing_from_api, grade_percent, passed_timestamp
        '''
        is_audit_enrollment = enterprise_enrollment.is_audit_enrollment
        lms_user_id = enterprise_enrollment.enterprise_customer_user.user_id
        enterprise_customer_uuid = enterprise_enrollment.enterprise_customer_user.enterprise_customer.uuid
        course_id = enterprise_enrollment.course_id

        if is_audit_enrollment:
            completed_date_from_api, grade_from_api, is_passing_from_api, grade_percent, passed_timestamp = \
                self.collect_grades_data(enterprise_enrollment, course_details, channel_name)
            if incomplete_count == 0 and completed_date_from_api is None:
                LOGGER.info(generate_formatted_log(
                    channel_name, enterprise_customer_uuid, lms_user_id, course_id,
                    'Setting completed_date to now() for audit course with all non-gated content done.'
                ))
                completed_date_from_api = timezone.now()
        else:
            completed_date_from_api, grade_from_api, is_passing_from_api, grade_percent, passed_timestamp = \
                self.collect_certificate_data(enterprise_enrollment, channel_name)
            LOGGER.info(generate_formatted_log(
                channel_name, enterprise_customer_uuid, lms_user_id, course_id,
                f'collect_certificate_data finished with CompletedDate: {completed_date_from_api},'
                f' Grade: {grade_from_api}, IsPassing: {is_passing_from_api},'
                f' Passed timestamp: {passed_timestamp}'
            ))
            if completed_date_from_api is None:
                # means we cannot find a cert for this learner
                # we will try getting grades info using the alternative api in this case
                # if that also does not exist then we have nothing to report
                completed_date_from_api, grade_from_api, is_passing_from_api, grade_percent, passed_timestamp = \
                    self.collect_grades_data(enterprise_enrollment, course_details, channel_name)
                LOGGER.info(generate_formatted_log(
                    channel_name, enterprise_customer_uuid, lms_user_id, course_id,
                    f'No certificate found, obtained grading data from grades api.'
                    f' CompletedDate: {completed_date_from_api},'
                    f' Grade: {grade_from_api}, IsPassing: {is_passing_from_api},'
                    f' Passed timestamp: {passed_timestamp}'
                ))

        # In the past we have been inconsistent about the format/source/typing of the grade_percent value.
        # Initial investigations have lead us to believe that grade percents from the source are seemingly more
        # accurate now, so we're enforcing float typing. To account for the chance that old documentation is
        # right and grade percents are reported as letter grade values, the float conversion is wrapped in a try/except
        # in order to both not blow up and also log said instance of letter grade values.
        try:
            if grade_percent is not None:
                grade_percent = float(grade_percent)
        except ValueError as exc:
            LOGGER.error(generate_formatted_log(
                channel_name, enterprise_customer_uuid, lms_user_id, course_id,
                f'Grade percent is not a valid float: {grade_percent}. Failed with exc: {exc}'
            ))
            grade_percent = None

        return completed_date_from_api, grade_from_api, is_passing_from_api, grade_percent, passed_timestamp

    def get_incomplete_content_count(self, enterprise_enrollment, channel_name):
        '''
        Fetch incomplete content count using completion blocks LMS api
        Will return None for non audit enrollment (but this does not have to be the case necessarily)
        '''
        incomplete_count = None
        is_audit_enrollment = enterprise_enrollment.is_audit_enrollment

        # The decision to not get incomplete count for non audit enrollments can be questioned
        # Right now we don't use this number for non audit. But this condition can be just removed
        # if we decide we need this for non audit enrollments too
        if not is_audit_enrollment:
            return incomplete_count
        lms_user_id = enterprise_enrollment.enterprise_customer_user.user_id
        enterprise_customer_uuid = enterprise_enrollment.enterprise_customer_user.enterprise_customer.uuid
        course_id = enterprise_enrollment.course_id

        user = User.objects.get(pk=lms_user_id)
        completion_summary = get_completion_summary(course_id, user)
        incomplete_count = completion_summary.get('incomplete_count')
        LOGGER.info(
            generate_formatted_log(
                channel_name, enterprise_customer_uuid, lms_user_id, course_id,
                f'Incomplete count for audit enrollment is {incomplete_count}'
            ))

        return incomplete_count

    def export(self, **kwargs):
        """
        Collect learner data for the ``EnterpriseCustomer`` where data sharing consent is granted.
        If BOTH learner_to_transmit and course_run_id are present, collected data returned is narrowed to
        that learner and course. If either param is absent or None, ALL data will be collected.

        Yields a learner data object for each enrollment, containing:

        * ``enterprise_enrollment``: ``EnterpriseCourseEnrollment`` object.
        * ``user_email``: PII User/learner email string
        * ``content_title``: Course title string
        * ``completed_date``: datetime instance containing the course/enrollment completion date; None if not complete.
          "Course completion" occurs for instructor-paced courses when course certificates are issued, and
          for self-paced courses, when the course end date is passed, or when the learner achieves a passing grade.
          Currently unused as it gets overridden from the grader data.
        * ``grade``: string grade recorded for the learner in the course.
        * ``learner_to_transmit``: OPTIONAL User, filters exported data
        * ``course_run_id``: OPTIONAL Course key string, filters exported data

        """
        channel_name = kwargs.get('app_label')
        lms_user_for_filter = kwargs.get('learner_to_transmit', None)
        course_run_id = kwargs.get('course_run_id', None)
        grade = kwargs.get('grade', None)
        skip_transmitted = kwargs.get('skip_transmitted', True)
        TransmissionAudit = kwargs.get('TransmissionAudit', None)

        # Fetch the consenting enrollment data, including the enterprise_customer_user.
        # Order by the course_id, to avoid fetching course API data more than we have to.
        enrollments_permitted = self._determine_enrollments_permitted(
            lms_user_for_filter,
            course_run_id,
            channel_name,
            skip_transmitted,
            TransmissionAudit,
            grade,
        )
        enrollment_ids_to_export = [enrollment.id for enrollment in enrollments_permitted]

        for enterprise_enrollment in enrollments_permitted:
            lms_user_id = enterprise_enrollment.enterprise_customer_user.user_id
            user_email = enterprise_enrollment.enterprise_customer_user.user_email
            enterprise_customer_uuid = enterprise_enrollment.enterprise_customer_user.enterprise_customer.uuid
            course_id = enterprise_enrollment.course_id

            course_details, error_message = LearnerExporterUtility.get_course_details_by_id(course_id)

            if course_details is None:
                # Course not found, so we have nothing to report.
                LOGGER.error(generate_formatted_log(
                    channel_name, enterprise_customer_uuid, lms_user_id, course_id,
                    f'get_course_details returned None for EnterpriseCourseEnrollment {enterprise_enrollment.pk}'
                    f', error_message: {error_message}'
                ))
                continue

            # For audit courses, check if 100% completed
            # which we define as: no non-gated content is remaining
            incomplete_count = self.get_incomplete_content_count(enterprise_enrollment, channel_name)

            (
                completed_date_from_api, grade_from_api,
                is_passing_from_api, grade_percent, passed_timestamp
            ) = self.get_grades_summary(
                course_details,
                enterprise_enrollment,
                channel_name,
                incomplete_count,
            )

            if completed_date_from_api:
                if is_passing_from_api:
                    progress_status = 'Passed'
                else:
                    progress_status = 'Failed'
            else:
                progress_status = 'In Progress'

            # Apply the Source of Truth for Grades
            # Note: Only completed records are transmitted by the completion transmitter
            #       therefore even non complete grading/cert records are exported here.
            records = self.get_learner_data_records(
                enterprise_enrollment=enterprise_enrollment,
                user_email=user_email,
                completed_date=completed_date_from_api,
                grade=grade_from_api,
                content_title=course_details.display_name,
                progress_status=progress_status,
                course_completed=is_course_completed(
                    enterprise_enrollment,
                    is_passing_from_api,
                    incomplete_count,
                    passed_timestamp,
                ),
                grade_percent=grade_percent,
            )

            if records:
                # There are some cases where we won't receive a record from the above
                # method; right now, that should only happen if we have an Enterprise-linked
                # user for the integrated channel, and transmission of that user's
                # data requires an upstream user identifier that we don't have (due to a
                # failure of SSO or similar). In such a case, `get_learner_data_record`
                # would return None, and we'd simply skip yielding it here.
                for record in records:
                    # Because we export a course and course run under the same enrollment, we can only remove the
                    # enrollment from the list of enrollments to export, once.
                    try:
                        enrollment_ids_to_export.pop(enrollment_ids_to_export.index(enterprise_enrollment.id))
                    except ValueError:
                        pass

                    yield record

        LOGGER.info(generate_formatted_log(
            channel_name, None, lms_user_for_filter, course_run_id,
            f'export finished. Did not export records for EnterpriseCourseEnrollment objects: '
            f' {enrollment_ids_to_export}.'
        ))

    def _filter_out_pre_transmitted_enrollments(
            self,
            enrollments_to_process,
            channel_name,
            grade,
            transmission_audit
    ):
        """
        Given an enrollments_to_process, returns only enrollments that are not already transmitted
        """
        included_enrollments = set()
        for enterprise_enrollment in enrollments_to_process:
            lms_user_id = enterprise_enrollment.enterprise_customer_user.user_id
            enterprise_customer_uuid = enterprise_enrollment.enterprise_customer_user.enterprise_customer.uuid
            course_id = enterprise_enrollment.course_id

            if transmission_audit and \
                    is_already_transmitted(
                        transmission_audit,
                        enterprise_enrollment.id,
                        self.enterprise_configuration.id,
                        grade,
                        detect_grade_updated=self.INCLUDE_GRADE_FOR_COMPLETION_AUDIT_CHECK,
                    ):
                # We've already sent a completion status for this enrollment
                LOGGER.info(generate_formatted_log(
                    channel_name, enterprise_customer_uuid, lms_user_id, course_id,
                    'Skipping export of previously sent enterprise enrollment. '
                    'EnterpriseCourseEnrollment: {enterprise_enrollment_id}'.format(
                        enterprise_enrollment_id=enterprise_enrollment.id
                    )))
                continue
            included_enrollments.add(enterprise_enrollment)
        return included_enrollments

    def get_enrollments_to_process(self, lms_user_for_filter, course_run_id, channel_name):
        """
        Fetches list of EnterpriseCourseEnrollments ordered by course_id.
        List is filtered by learner and course_run_id if both are provided

        lms_user_for_filter: If None, data for ALL courses and learners will be returned
        course_run_id: If None, data for ALL courses and learners will be returned

        """
        enrollment_queryset = EnterpriseCourseEnrollment.objects.select_related(
            'enterprise_customer_user'
        ).filter(
            enterprise_customer_user__enterprise_customer=self.enterprise_customer,
            enterprise_customer_user__active=True,
        )
        if lms_user_for_filter and course_run_id:
            enrollment_queryset = enrollment_queryset.filter(
                course_id=course_run_id,
                enterprise_customer_user__user_id=lms_user_for_filter.id,
            )
            LOGGER.info(generate_formatted_log(
                channel_name, self.enterprise_customer.uuid, lms_user_for_filter, course_run_id,
                'get_enrollments_to_process run for single learner and course.'))
        enrollment_queryset = enrollment_queryset.order_by('course_id')
        # return resolved list instead of queryset
        return list(enrollment_queryset)

    def get_learner_assessment_data_records(
            self,
            enterprise_enrollment,
            assessment_grade_data
    ):
        """
        Generate a learner assessment data transmission audit with fields properly filled in.
        Returns a list of LearnerDataTransmissionAudit objects.

        enterprise_enrollment: the ``EnterpriseCourseEnrollment`` object we are getting a learner's data for.
        assessment_grade_data: A dict with keys corresponding to different edX course subsections.
        See _collect_assessment_grades_data for the formatted data returned as the value for a given key.
        """
        TransmissionAudit = apps.get_model('integrated_channel', 'GenericLearnerDataTransmissionAudit')
        user_subsection_audits = []
        # Create an audit for each of the subsections in the course data.
        for subsection_data in assessment_grade_data.values():
            subsection_percent_grade = subsection_data.get('grade')
            subsection_id = subsection_data.get('subsection_id')
            # Sanity check for a grade to report
            if not subsection_percent_grade or not subsection_id:
                continue

            user_subsection_audits.append(TransmissionAudit(
                plugin_configuration_id=self.enterprise_configuration.id,
                enterprise_customer_uuid=self.enterprise_configuration.enterprise_customer.uuid,
                enterprise_course_enrollment_id=enterprise_enrollment.id,
                course_id=enterprise_enrollment.course_id,
                subsection_id=subsection_id,
                grade=subsection_percent_grade,
            ))

        return user_subsection_audits

    def get_learner_data_records(
            self,
            enterprise_enrollment,
            completed_date=None,
            grade=None,
            course_completed=False,
            grade_percent=None,
            content_title=None,
            user_email=None,
            progress_status=None,

    ):  # pylint: disable=unused-argument
        """
        Generate a learner data transmission audit with fields properly filled in.
        """
        TransmissionAudit = apps.get_model('integrated_channel', 'GenericLearnerDataTransmissionAudit')
        completed_timestamp = None
        if completed_date is not None:
            completed_timestamp = parse_datetime_to_epoch_millis(completed_date)
        # We return two records here, one with the course key and one with the course run id, to account for
        # uncertainty about the type of content (course vs. course run) that was sent to the integrated channel.
        return [
            TransmissionAudit(
                plugin_configuration_id=self.enterprise_configuration.id,
                enterprise_customer_uuid=self.enterprise_configuration.enterprise_customer.uuid,
                enterprise_course_enrollment_id=enterprise_enrollment.id,
                course_id=get_course_id_for_enrollment(enterprise_enrollment),
                course_completed=course_completed,
                completed_timestamp=completed_timestamp,
                grade=grade,
                user_email=user_email,
                content_title=content_title,
                progress_status=progress_status,
            ),
            TransmissionAudit(
                plugin_configuration_id=self.enterprise_configuration.id,
                enterprise_customer_uuid=self.enterprise_configuration.enterprise_customer.uuid,
                enterprise_course_enrollment_id=enterprise_enrollment.id,
                course_id=enterprise_enrollment.course_id,
                course_completed=course_completed,
                completed_timestamp=completed_timestamp,
                grade=grade,
                user_email=user_email,
                content_title=content_title,
                progress_status=progress_status,
            )
        ]

    def collect_certificate_data(self, enterprise_enrollment, channel_name):
        """
        Collect the learner completion data from the course certificate.

        Used for Instructor-paced courses.

        If no certificate is found, then returns the completed_date = None, grade = In Progress, on the idea that a
        certificate will eventually be generated.

        Args:
            enterprise_enrollment (EnterpriseCourseEnrollment): the enterprise enrollment record for which we need to
            collect completion/grade data,
            channel_name: labeled for relevant integrated channel this is being called for to enhance logging.

        Returns:
            completed_date: Date the course was completed, this is None if course has not been completed.
            grade: Current grade in the course.
            percent_grade: The current percent grade in the course.
            is_passing: Boolean indicating if the grade is a passing grade or not.
            passed_timestamp: Timestamp when learner obtained passing grade.
        """

        course_id = enterprise_enrollment.course_id
        lms_user_id = enterprise_enrollment.enterprise_customer_user.user_id
        enterprise_customer_uuid = enterprise_enrollment.enterprise_customer_user.enterprise_customer.uuid
        user = User.objects.get(pk=lms_user_id)
        passed_timestamp = None

        completed_date = None
        grade = self.grade_incomplete
        is_passing = False
        percent_grade = None

        try:
            certificate = get_course_certificate(course_id, user)
        except InvalidKeyError:
            certificate = None
            LOGGER.error(generate_formatted_log(
                channel_name, enterprise_customer_uuid, lms_user_id, course_id,
                'get_course_certificate failed. Certificate fetch failed due to invalid course_id for'
                f' EnterpriseCourseEnrollment: {enterprise_enrollment}. Data export will continue without grade.'
            ))

        if not certificate:
            return completed_date, grade, is_passing, percent_grade, passed_timestamp

        LOGGER.info(generate_formatted_log(
            channel_name, enterprise_customer_uuid, lms_user_id, course_id,
            f'get_course_certificate certificate={certificate}'
        ))

        # get_certificate_for_user has an optional formatting argument which defaults to True
        # edx-platform/blob/6b9eb122dd5b018cfeffc120a70e503b2c159c0b/lms/djangoapps/certificates/api.py#L128
        # when True, the certifiate's `created_date` is transformed into just `created`
        completed_date = certificate.get('created', None)
        # guard against the default formatting argument changing
        if not completed_date:
            completed_date = certificate.get('created_date', None)
        if completed_date:
            completed_date = parse_datetime(str(completed_date))

        # also get passed_timestamp which is used to line up completion logic with analytics
        persistent_grade = get_persistent_grade(course_id, user)
        passed_timestamp = persistent_grade.passed_timestamp if persistent_grade is not None else None

        if (not completed_date) and passed_timestamp:
            LOGGER.info(generate_formatted_log(
                channel_name, enterprise_customer_uuid, lms_user_id, course_id,
                'get_course_certificate misisng created/created_date, '
                'but there is a passed_timestamp so using that'
            ))
            completed_date = datetime.fromtimestamp((passed_timestamp / 1000), tz=timezone.utc)
        elif not completed_date and not passed_timestamp:
            LOGGER.info(generate_formatted_log(
                channel_name, enterprise_customer_uuid, lms_user_id, course_id,
                'get_course_certificate misisng created/created_date, '
                'no passed_timestamp so defaulting to timezone.now(), '
                f'certificate={certificate}'
            ))
            completed_date = timezone.now()

        # For consistency with _collect_grades_data, we only care about Pass/Fail grades. This could change.
        is_passing = certificate.get('is_passing')
        percent_grade = certificate.get('grade')
        grade = self.grade_passing if is_passing else self.grade_failing

        return completed_date, grade, is_passing, percent_grade, passed_timestamp

    def _collect_assessment_grades_data(self, enterprise_enrollment):
        """
        Collect a learner's assessment level grade data using an enterprise enrollment, from the Grades API.

        Args:
            enterprise_enrollment (EnterpriseCourseEnrollment): the enterprise enrollment record for which we need to
            collect subsection grades data
        Returns:
            Dict:
                {
                    [subsection name]: {
                        'grade_category': category,
                        'grade': percent grade,
                        'assessment_label': label,
                        'grade_point_score': points earned on the assignment,
                        'grade_points_possible': max possible points on the assignment,
                        'subsection_id': subsection module ID
                    }

                    ...
                }
        """
        if self.grades_api is None:
            self.grades_api = GradesApiClient(self.user)

        course_id = enterprise_enrollment.course_id
        username = enterprise_enrollment.enterprise_customer_user.user.username
        try:
            assessment_grades_data = self.grades_api.get_course_assessment_grades(course_id, username)
        except HTTPError as err:
            if err.response and err.response.status_code == 404:
                return {}
            raise err

        assessment_grades = {}
        for grade in assessment_grades_data:
            if not grade.get('attempted'):
                continue
            assessment_grades[grade.get('subsection_name')] = {
                'grade_category': grade.get('category'),
                'grade': grade.get('percent'),
                'assessment_label': grade.get('label'),
                'grade_point_score': grade.get('score_earned'),
                'grade_points_possible': grade.get('score_possible'),
                'subsection_id': grade.get('module_id')
            }

        return assessment_grades

    def collect_grades_data(self, enterprise_enrollment, course_details, channel_name):
        """
        Collect the learner completion data from the Grades API.

        Used for self-paced courses.

        Args:
            enterprise_enrollment (EnterpriseCourseEnrollment): the enterprise enrollment record for which we need to
            collect completion/grade data
            course_details (CourseOverview): the course details for the course in the enterprise enrollment record.
            channel_name: Integrated channel name for improved logging.

        Returns:
            completed_date: Date the course was completed, None if course has not been completed.
            grade: Current grade in the course.
            is_passing: Boolean indicating if the grade is a passing grade or not.
            percent_grade: a number between 0 and 100
            passed_timestamp: Timestamp when learner obtained passing grade.
        """

        course_id = enterprise_enrollment.course_id
        lms_user_id = enterprise_enrollment.enterprise_customer_user.user_id
        user = User.objects.get(pk=lms_user_id)
        enterprise_customer_uuid = enterprise_enrollment.enterprise_customer_user.enterprise_customer.uuid

        grades_data = get_single_user_grade(course_id, user)

        if grades_data is None:
            LOGGER.warning(generate_formatted_log(
                channel_name, enterprise_customer_uuid, lms_user_id, course_id,
                f'No grade found for '
                f'EnterpriseCourseEnrollment: {enterprise_enrollment}.'
            ))
            # if enrollment found, but no grades, we can safely mark as incomplete/in progress
            return None, LearnerExporter.GRADE_INCOMPLETE, None, None, None

        # also get passed_timestamp which is used to line up completion logic with analytics
        persistent_grade = get_persistent_grade(course_id, user)
        passed_timestamp = persistent_grade.passed_timestamp if persistent_grade is not None else None

        # Prepare to process the course end date and pass/fail grade
        course_end_date = course_details.end
        now = timezone.now()
        is_passing = grades_data.passed

        # We can consider a course complete if:
        # * the course's end date has passed
        if course_end_date is not None and course_end_date < now:
            completed_date = course_end_date
            grade = self.grade_passing if is_passing else self.grade_failing
            grade = self.grade_audit if enterprise_enrollment.is_audit_enrollment else grade

        # * Or, the learner has a passing grade (as of now)
        elif is_passing:
            completed_date = now
            grade = self.grade_passing

        # Otherwise, the course is still in progress
        else:
            completed_date = None
            grade = self.grade_incomplete

        percent_grade = grades_data.percent

        return completed_date, grade, is_passing, percent_grade, passed_timestamp


class LearnerExporterUtility:
    """ Utility class namespace for accessing Django objects in a common way. """

    @staticmethod
    def lms_user_id_for_ent_course_enrollment_id(enterprise_course_enrollment_id):
        """ Returns the ID of the LMS User for the EnterpriseCourseEnrollment id passed in
        or None if EnterpriseCourseEnrollment not found """
        try:
            return EnterpriseCourseEnrollment.objects.get(
                id=enterprise_course_enrollment_id).enterprise_customer_user.user_id
        except EnterpriseCourseEnrollment.DoesNotExist:
            return None

    @staticmethod
    def get_course_details_by_id(course_id):
        '''
        Convenience method to fetch course details or None (if not found)
        '''
        course_details = None
        error_message = None
        try:
            course_details = get_course_details(course_id)
        except (InvalidKeyError, CourseOverview.DoesNotExist):
            error_message = f'get_course_details failed for course_id {course_id}'

        return course_details, error_message
